from __future__ import annotations

import functools
import inspect
from typing import TYPE_CHECKING, Union

from oarepo_model_builder.utils.json_pathlib import JSONPaths
from oarepo_model_builder.stack import ModelBuilderStack, ReplaceElement
from oarepo_model_builder.utils.verbose import log

if TYPE_CHECKING:
    from oarepo_model_builder.builder import ModelBuilder


def process(path, priority=0, condition=None):
    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        wrapped.model_builder_path = path
        wrapped.model_builder_priority = priority
        wrapped.model_builder_condition = condition
        return wrapped

    return wrapper


class OutputBuilder:
    TYPE = None

    def __init__(self, builder: ModelBuilder):
        self.builder = builder
        # TODO: move this to metaclass and initialize only once per class
        self.json_paths = JSONPaths()
        arr = []
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if not hasattr(method, 'model_builder_priority'):
                continue
            arr.append(
                (
                    -method.model_builder_priority,
                    -len(method.model_builder_path),
                    method.model_builder_path,
                    id(method),
                    method.model_builder_condition,
                    method
                )
            )
        arr.sort()
        for _prior, _lpath, path, _mid, condition, method in arr:
            self.json_paths.register(path, condition, method)

    def begin(self, schema, settings):
        self.schema = schema
        self.settings = settings
        log.enter(2, 'Creating %s', self.TYPE)
        pass

    def finish(self):
        log.leave()

    def process_element(self, stack: ModelBuilderStack):
        """
        Normally returns a generator with a single yield:
        1. first part is called on element start
        2. yield causes that the content of the element is processed
        3. generator is called again to finish the element

        If no generator is returned, the content of the element is not processed
        """
        try:
            self.call_components('before_process_element', value=stack.top.data, stack=stack)
            for method in self.json_paths.match(stack.path, stack.top.data, extra_data={
                'stack': stack
            }):
                return method(stack)
            # do not skip stack top
            if stack.level > 1:
                return stack.SKIP
        finally:
            self.call_components('after_process_element', value=stack.top.data, stack=stack)

    @process('/model')
    def enter_model(self, stack: ModelBuilderStack):
        # do not skip /model
        yield

    def call_components(self, method_name, value, **kwargs):
        for component in self.builder.get_output_builder_components(self.TYPE):
            if hasattr(component, method_name):
                value = getattr(component, method_name)(self, value, **kwargs) or value
        return value


class OutputBuilderComponent:
    def before_process_element(self, builder: OutputBuilder, value, *, stack: ModelBuilderStack, **kwargs):
        return value

    def after_process_element(self, builder: OutputBuilder, value, *, stack: ModelBuilderStack, **kwargs):
        return value


__all__ = [
    'process',
    'OutputBuilder',
    'ModelBuilderStack',
    'ReplaceElement'
]



from libcst import CSTTransformer, ClassDef, FunctionDef, Module, SimpleStatementLine, Import, ImportFrom, Assign, Name, \
    AssignTarget, List


class MergingTransformer(CSTTransformer):
    def __init__(self, new_cst):
        super().__init__()
        self.new_cst = new_cst
        self.stack = []
        self._push(new_cst.body)

    def _push(self, new_members):
        self.stack.append((new_members, {}))

    def _pop(self):
        return self.stack.pop()

    def _set(self, key, value):
        self.stack[-1][1][key] = value

    def visit_ClassDef(self, node: ClassDef):
        self._set(node.name.value, node)
        for part in self.stack[-1][0]:
            if hasattr(part, 'name'):
                if part.name.value == node.name.value:
                    if hasattr(part, 'body'):
                        self._push(part.body.body)
                        break
        else:
            self._push([])

    def visit_FunctionDef(self, node: FunctionDef):
        # do not visit the children of the function as we do not modify
        # an already written function code
        self._set(node.name.value, node)
        return False

    def leave_ClassDef(
            self, original_node: ClassDef, updated_node: ClassDef
    ):
        parts_to_add = self.leave_class_module()
        self._pop()

        if parts_to_add:
            # and merge them with the original content
            return updated_node.with_changes(
                body=updated_node.body.with_changes(
                    body=[
                        *updated_node.body.body,
                        *parts_to_add
                    ]
                )
            )
        return updated_node

    def visit_Assign(self, node):
        assign_name = self._node_name(node)
        stack_top = self.stack[-1][0]
        for el in stack_top:
            el_name = self._node_name(el)
            if el_name == assign_name and isinstance(node.value, List):
                self._push(remove_simple_line(el))
                return node
        self._push(None)
        return False

    def leave_Assign(self, original_node, updated_node):
        self._pop()
        return updated_node

    def leave_List(self, original_node, updated_node):
        stack_top = self.stack[-1][0]
        if not isinstance(stack_top, Assign):
            return False
        # TODO: merge
        node_elements = list(updated_node.elements)
        for el in stack_top.value.elements:
            for ne in updated_node.elements:
                if ne.value.deep_equals(el.value):
                    break
            else:
                node_elements.append(el)
        return updated_node.with_changes(
            elements=node_elements
        )


    def _merge_assign_value(self, target_node, source_node):
        return target_node

    def leave_Module(self, original_node: Module, updated_node: Module):
        parts_to_add = self.leave_class_module()

        extra_imports = self._merge_imports(original_node)
        extra_assigns = self._merge_assigns(original_node)

        self._pop()

        # and merge
        return updated_node.with_changes(
            body=[
                *extra_imports,
                *updated_node.body,
                *parts_to_add,
                *extra_assigns
            ]
        )

    def _merge_imports(self, original_node):
        new_imports = self._extract_imports(self.stack[0][0])
        existing_imports = self._extract_imports(original_node.body)
        return self._merge_nodes(existing_imports, new_imports)

    def _merge_assigns(self, original_node):
        new_assigns = self._extract_assigns(self.stack[0][0])
        existing_assigns = self._extract_assigns(original_node.body)
        return self._merge_nodes(existing_assigns, new_assigns)

    def _merge_nodes(self, existing_nodes, new_nodes):
        # extract nodes that are not yet present
        extra_nodes = []
        for ni in new_nodes:
            for ei in existing_nodes:
                if self._node_name(ei) == self._node_name(ni):
                    break
            else:
                extra_nodes.append(ni)
        return extra_nodes

    def _node_name(self, node):
        if isinstance(node, SimpleStatementLine):
            body = node.body
            if not len(body):
                raise NotImplementedError(f'No body in SimpleStatementLine {node}')
            name_element = body[0]
        else:
            name_element = node

        if isinstance(name_element, Assign):
            lhs = name_element.children[0]
            if not isinstance(lhs, AssignTarget):
                raise NotImplementedError(f'Getting name of node type {type(lhs)} not implemented yet')
            return lhs.target.value
        elif isinstance(name_element, (Import, ImportFrom)):
            return self.new_cst.code_for_node(name_element).strip()
        else:
            raise NotImplementedError(f'Getting name of node type {type(name_element)} not implemented yet')

    def _extract_imports(self, lines):
        return self._extract_top_level(lines, (Import, ImportFrom))

    def _extract_assigns(self, lines):
        return self._extract_top_level(lines, Assign)

    def _extract_top_level(self, lines, classes=()):
        extracted = []
        for l in lines:
            if isinstance(l, SimpleStatementLine):
                for ml in l.children:
                    if isinstance(ml, classes):
                        extracted.append(l)
                        break
        return extracted

    def leave_class_module(self):
        new_parts = self.stack[-1][0]
        existing_parts = self.stack[-1][1]

        # look for the parts in the new parts that are not in the existing parts
        parts_to_add = []
        for part in new_parts:
            if hasattr(part, 'name'):
                if part.name.value not in existing_parts:
                    parts_to_add.append(part)
        return parts_to_add


def remove_simple_line(node):
    if isinstance(node, SimpleStatementLine):
        return node.body[0]
    return node



import sys
from pathlib import Path

import pkg_resources

from oarepo_model_builder.builder import ModelBuilder
from oarepo_model_builder.schema import ModelSchema
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch


def create_builder_from_entrypoints(**kwargs):
    output_classes = load_entry_points_list('oarepo_model_builder.ouptuts')
    builder_classes = load_entry_points_list('oarepo_model_builder.builders')
    preprocess_classes = load_entry_points_list('oarepo_model_builder.property_preprocessors')
    model_preprocessor_classes = load_entry_points_list('oarepo_model_builder.model_preprocessors')

    builder_types = [x.TYPE for x in builder_classes]
    output_builder_components = {
        builder_type: load_entry_points_list(f'oarepo_model_builder.builder_components.{builder_type}')
        for builder_type in builder_types
    }

    return ModelBuilder(
        output_builders=builder_classes,
        outputs=output_classes,
        property_preprocessors=preprocess_classes,
        model_preprocessors=model_preprocessor_classes,
        output_builder_components=output_builder_components,
        **kwargs
    )


def load_entry_points_dict(name):
    return {ep.name: ep.load() for ep in pkg_resources.iter_entry_points(group=name)}


def load_entry_points_list(name):
    ret = [(ep.name, ep.load()) for ep in pkg_resources.iter_entry_points(group=name)]
    ret.sort()
    return [x[1] for x in ret]


def load_model_from_entrypoint(schema, ep: pkg_resources.EntryPoint):
    filename = '.'.join(ep.attrs)
    data = pkg_resources.resource_string(ep.module_name, filename)
    return schema._load(filename, content=data)


def load_included_models_from_entry_points():
    ret = {}
    for ep in pkg_resources.iter_entry_points(group='oarepo.models'):
        ret[ep.name] = lambda schema: load_model_from_entrypoint(schema, ep)
    return ret


def load_model(model_filename, package=None, configs=(), black=True, isort=True, sets=(), model_content=None):
    loaders = load_entry_points_dict('oarepo_model_builder.loaders')
    included_models = load_included_models_from_entry_points()
    schema = ModelSchema(model_filename, content=model_content,
                         loaders=loaders, included_models=included_models)
    for config in configs:
        load_config(schema, config, loaders)
    for s in sets:
        k, v = s.split('=', 1)
        schema.schema[k] = v
    check_plugin_packages(schema)
    if package:
        schema.settings['package'] = package
    if 'python' not in schema.settings:
        schema.settings.python = HyphenMunch()
    schema.settings.python.use_isort = isort
    schema.settings.python.use_black = black
    return schema


def load_config(schema, config, loaders):
    old_loaders = schema.loaders
    schema.loaders = loaders
    try:
        loaded_file = schema._load(config)
        schema.merge(loaded_file)
    finally:
        schema.loaders = old_loaders


def check_plugin_packages(schema):
    try:
        required_packages = schema.schema.plugins.packages
    except AttributeError:
        return
    import pkg_resources, subprocess
    known_packages = set(d.project_name for d in pkg_resources.working_set)
    unknown_packages = [rp for rp in required_packages if rp not in known_packages]
    if unknown_packages:
        if input(f'Required packages {", ".join(unknown_packages)} are missing. '
                 f'Should I install them for you via pip install? (y/n) ') == 'y':
            if subprocess.call([
                'pip', 'install', *unknown_packages
            ]):
                sys.exit(1)
            print("Installed required packages, please run this command again")
        sys.exit(1)



from ..builders import OutputBuilder
from ..outputs.toml import TOMLOutput
from ..utils.verbose import log


class InvenioSampleAppPoetryBuilder(OutputBuilder):
    TYPE = 'invenio_sample_app_poetry'

    def finish(self):
        super().finish()

        output: TOMLOutput = self.builder.get_output(
            'toml',
            'pyproject.toml'
        )

        output.setdefault("tool.poetry.dependencies.invenio",
                          'version', '^3.5.0a1',
                          'extras', ["base", "auth", "metadata", "files", "postgresql", "elasticsearch7"],
                          'optional', True,
                          'allow-prereleases', True
                          )

        output.setdefault("tool.poetry.dependencies.invenio-records-resources",
                          'version', '^0.17.3',
                          'optional', True,
                          'allow-prereleases', True
                          )

        output.setdefault("tool.poetry.extras", 'sample-app', ['invenio', 'invenio-records-resources'])

        if output.created:
            log(log.INFO, f"""To install the sample app, run
    poetry install -E sample-app            
            """)



from oarepo_model_builder.utils.deepmerge import deepmerge
from oarepo_model_builder.model_preprocessors import ModelPreprocessor
from oarepo_model_builder.utils.camelcase import camel_case, snake_case


def last_item(x):
    return x.rsplit('.', maxsplit=1)[-1]


class InvenioModelPreprocessor(ModelPreprocessor):
    TYPE = 'invenio'

    def transform(self, schema, settings):
        deepmerge(settings, {
            'python': {
                'record_prefix': camel_case(settings.package.rsplit('.', maxsplit=1)[-1]),
                # just make sure that the templates is always there
                'templates': {
                },
                'marshmallow': {
                    'mapping': {
                    }
                }
            }
        })

        settings.setdefault('top-level-metadata', True)

        record_prefix = settings.python.record_prefix
        self.set(settings.python, 'record-prefix-snake',
                 lambda: snake_case(settings.python.record_prefix))

        # config
        self.set(settings.python, 'config-package',
                 lambda: f'{settings.package}.config')
        self.set(settings.python, 'config-dummy-class',
                 lambda: f'{settings.package}.config.DummyClass')
        self.set(settings.python, 'config-resource-config-key',
                 lambda: f'{settings.package_base_upper}_RESOURCE_CONFIG')
        self.set(settings.python, 'config-resource-class-key',
                 lambda: f'{settings.package_base_upper}_RESOURCE_CLASS')
        self.set(settings.python, 'config-service-config-key',
                 lambda: f'{settings.package_base_upper}_SERVICE_CONFIG')
        self.set(settings.python, 'config-service-class-key',
                 lambda: f'{settings.package_base_upper}_SERVICE_CLASS')

        # ext
        self.set(settings.python, 'ext-class',
                 lambda: f'{settings.package}.ext.{record_prefix}Ext')
        self.set(settings.python, 'flask-extension-name',
                 lambda: f'{settings.package_base}')

        # proxies
        self.set(settings.python, 'proxies-current-resource',
                 lambda: f'{settings.package}.proxies.current_resource')

        # record
        self.set(settings.python, 'record-class',
                 lambda: f'{settings.package}.records.api.{record_prefix}Record')
        self.set(settings.python, 'record-metadata-class',
                 lambda: f'{settings.package}.records.models.{record_prefix}Metadata')
        self.set(settings.python, 'record-metadata-table-name',
                 lambda: f'{record_prefix.lower()}_metadata')
        #   - poetry
        self.set(settings.python, 'record-mapping-poetry',
                 lambda: f'{settings.package_base}')
        self.set(settings.python, 'record-jsonschemas-poetry',
                 lambda: f'{settings.package_base}')

        # resource
        self.set(settings.python, 'record-resource-config-class',
                 lambda: f'{settings.package}.resources.config.{record_prefix}ResourceConfig')
        self.set(settings.python, 'record-resource-class',
                 lambda: f'{settings.package}.resources.resource.{record_prefix}Resource')
        self.set(settings.python, 'record-permissions-class',
                 lambda: f'{settings.package}.resources.permissions.{record_prefix}PermissionPolicy')

        # service
        self.set(settings.python, 'record-service-class',
                 lambda: f'{settings.package}.services.service.{record_prefix}Service')
        self.set(settings.python, 'record-service-config-class',
                 lambda: f'{settings.package}.services.config.{record_prefix}ServiceConfig')
        #   - schema
        self.set(settings.python, 'record-schema-class',
                 lambda: f'{settings.package}.services.schema.{record_prefix}Schema')
        self.set(settings.python, 'record-schema-metadata-class',
                 lambda: f'{settings.package}.services.schema.{record_prefix}MetadataSchema')
        #   - dumper
        self.set(settings.python, 'record-dumper-class',
                 lambda: f'{settings.package}.services.dumper.{record_prefix}Dumper')
        #   - search
        self.set(settings.python, 'record-search-options-class',
                 lambda: f'{settings.package}.services.search.{record_prefix}SearchOptions')

        # alembic
        self.set(settings.python, 'record-schema-metadata-alembic',
                 lambda: f'{settings.package_base}')
        self.set(settings.python, 'record-schema-metadata-poetry',
                 lambda: f'{settings.package_base}')

        self.set(settings.python, 'record-resource-blueprint-name',
                 lambda: record_prefix)
        self.set(settings.python, 'create-blueprint-from-app',
                 lambda: f'{settings.package}.views.create_blueprint_from_app')

        if 'model' in schema.schema:
            if settings.top_level_metadata:
                schema_class = settings.python.record_schema_class
                schema_class_base_classes = settings.python.get('record_schema_bases', []) + [
                    'ma.Schema'  # alias will be recognized automatically
                ]
            else:
                schema_class = settings.python.record_schema_metadata_class,
                schema_class_base_classes = settings.python.get('record_schema_metadata_bases', []) + [
                    'ma.Schema'  # alias will be recognized automatically
                ]

            deepmerge(
                schema.schema.model.setdefault('oarepo:marshmallow', {}),
                {
                    'class': schema_class,
                    'base-classes': schema_class_base_classes,
                    'generate': True
                })

        # default import prefixes
        settings.python.setdefault('always-defined-import-prefixes', []).extend(
            ['ma', 'ma_fields', 'ma_valid']
        )



import json

from deepdiff import DeepDiff

from . import OutputBase
from .json_stack import JSONStack
from ..utils.verbose import log

try:
    import json5
except ImportError:
    import json as json5


class JSONOutput(OutputBase):
    IGNORE_NODE = JSONStack.IGNORED_NODE
    IGNORE_SUBTREE = JSONStack.IGNORED_SUBTREE

    def begin(self):
        try:
            with self.builder.filesystem.open(self.path) as f:
                self.original_data = json5.load(f)  # noqa
        except FileNotFoundError:
            self.original_data = None
        except ValueError:
            self.original_data = None

        self.stack = JSONStack()

    @property
    def created(self):
        return self.original_data is None

    def finish(self):
        data = self.stack.value
        if DeepDiff(data, self.original_data):
            self.builder.filesystem.mkdir(self.path.parent)
            log(2, 'Saving %s', self.path)
            with self.builder.filesystem.open(self.path, mode='w') as f:
                json.dump(data, f, ensure_ascii=False, indent=4)

    def enter(self, key, el):
        if key:
            self.stack.push(key, el)

    def leave(self):
        if not self.stack.empty:
            self.stack.pop()

    def primitive(self, key, value):
        if key:
            self.stack.push(key, value)
            self.stack.pop()

    def merge(self, value):
        self.stack.merge(value)


from __future__ import annotations

import functools
import inspect
from collections import namedtuple
from typing import TYPE_CHECKING

from oarepo_model_builder.utils.json_pathlib import JSONPaths
from oarepo_model_builder.stack import ModelBuilderStack

if TYPE_CHECKING:
    from oarepo_model_builder.builder import ModelBuilder


class PropertyPreprocessor:
    PathMethodRecord = namedtuple('PathMethodRecord', 'method, output_builder_type')

    def __init__(self, builder: ModelBuilder):
        self.builder = builder
        # TODO: move this to metaclass and initialize only once per class
        self.json_paths = JSONPaths()
        arr = []
        for name, method in inspect.getmembers(self, inspect.ismethod):
            if not hasattr(method, 'model_builder_priority'):
                continue
            arr.append(
                (
                    -method.model_builder_priority,
                    -len(method.model_builder_path),
                    method.model_builder_path,
                    id(method),
                    method.model_builder_condition,
                    method.model_builder_output_builder_type,
                    method
                )
            )
        arr.sort()
        for _prior, _lpath, path, _mid, condition, output_builder_type, method in arr:
            self.json_paths.register(path, condition,
                                     PropertyPreprocessor.PathMethodRecord(method, output_builder_type))

    def begin(self, schema, settings):
        self.schema = schema
        self.settings = settings

    def finish(self):
        pass

    def _call_method(self, data, stack: ModelBuilderStack, output_builder_type):
        for method, _output_builder_type in self.json_paths.match(
                stack.path, stack.top.data,
                extra_data={
                    'stack': stack
                }):
            if _output_builder_type == '*' or output_builder_type == _output_builder_type:
                return method(data, stack=stack)

    def process(self, output_builder_type: str, data, stack: ModelBuilderStack):
        return self._call_method(data, stack, output_builder_type)


def process(model_builder, path, priority=0, condition=None):
    def wrapper(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        wrapped.model_builder_priority = priority
        wrapped.model_builder_output_builder_type = \
            model_builder if isinstance(model_builder, str) else model_builder.TYPE
        wrapped.model_builder_path = path
        wrapped.model_builder_condition = condition
        return wrapped

    return wrapper



from pathlib import Path

from oarepo_model_builder.stack import ModelBuilderStack
from . import process
from .json_base import JSONBaseBuilder
from .utils import ensure_parent_modules
from ..utils.schema import is_schema_element


class JSONSchemaBuilder(JSONBaseBuilder):
    TYPE = 'jsonschema'
    output_file_type = 'jsonschema'
    output_file_name = 'schema-file'
    parent_module_root_name = 'jsonschemas'

    @process('/model/**', condition=lambda current, stack: is_schema_element(stack))
    def model_element(self, stack: ModelBuilderStack):
        self.model_element_enter(stack)
        yield
        self.model_element_leave(stack)

    def on_enter_model(self, output_name, stack: ModelBuilderStack):
        ensure_parent_modules(self.builder, Path(output_name),
                              ends_at=self.parent_module_root_name)



from oarepo_model_builder.builders import OutputBuilder, process
from oarepo_model_builder.stack import ModelBuilderStack


class JSONBaseBuilder(OutputBuilder):
    output_file_name: str = None
    output_file_type: str = None
    parent_module_root_name: str = None

    def model_element_enter(self, stack: ModelBuilderStack):
        top = stack.top
        data = top.data
        data = self.call_components('model_element_enter', data, stack=stack)
        match stack.top_type:
            case stack.PRIMITIVE:
                self.output.primitive(top.key, data)
            case stack.LIST:
                self.output.enter(top.key, [])
            case stack.DICT:
                self.output.enter(top.key, {})

    def model_element_leave(self, stack: ModelBuilderStack):
        self.call_components('model_element_leave', stack.top.data, stack=stack)
        if stack.top_type != stack.PRIMITIVE:
            self.output.leave()

    @process('/model')
    def enter_model(self, stack: ModelBuilderStack):
        output_name = self.settings[self.output_file_name]
        self.output = self.builder.get_output(self.output_file_type, output_name)
        self.on_enter_model(output_name, stack)

    def on_enter_model(self, output_name, stack: ModelBuilderStack):
        pass



from functools import cached_property
from typing import Generator, List

from deepdiff import DeepDiff


class ReplaceElement(Exception):
    def __init__(self, data):
        super().__init__()
        self.data = data


class ModelBuilderStackEntry:
    def __init__(self, key=None, data=None):
        self.key = key
        self.data = data

    def __getitem__(self, item):
        return self.data[item]

    def __eq__(self, other):
        return self.key == other.key and not DeepDiff(self.key, other.key)

    def __str__(self):
        return f'{self.key} - {self.data}'


class ModelBuilderStack:
    DICT = 'dict'
    LIST = 'list'
    PRIMITIVE = 'primitive'
    SKIP = 'skip'

    def __init__(self, schema):
        self.schema = schema
        self.stack = []

    def __getitem__(self, item):
        return self.stack[item]

    def push(self, key, el):
        self._clear_path()
        self.stack.append(ModelBuilderStackEntry(key, el))

    def pop(self):
        self._clear_path()
        self.stack.pop()

    @property
    def top(self):
        return self.stack[-1]

    @property
    def level(self):
        return len(self.stack)

    @property
    def top_type(self):
        match self.top.data:
            case dict():
                return self.DICT
            case list():
                return self.LIST
            case _:
                return self.PRIMITIVE

    @cached_property
    def path(self):
        return '/' + '/'.join(x.key for x in self.stack if x.key)

    def _clear_path(self):
        if "path" in self.__dict__:
            del self.__dict__["path"]

    def process(self, on_element):
        self.stack = []
        try:
            processing_order = self.schema.schema.processing_order
        except AttributeError:
            processing_order = None
        self._process_internal(None, self.schema.schema, on_element, processing_order)

    def _process_internal(self, key, element, on_element, processing_order: List[str] = None):
        popped = False

        try:
            # push the element to the stack
            self.push(key, element)

            # call the on_element function.
            ret = on_element(self)

            # if the result is not a generator,
            if not isinstance(ret, Generator):
                ret = iter([ret])

            res = next(ret, '')

            if res is self.SKIP:
                return

            match self.top_type:
                case self.LIST:
                    for idx, l in enumerate(self.top.data):
                        self._process_internal(idx, l, on_element)
                case self.DICT:
                    items = list(self.top.data.items())
                    if processing_order:
                        def key_function(x):
                            try:
                                return processing_order.index(x)
                            except ValueError:
                                pass
                            try:
                                return processing_order.index('*')
                            except ValueError:
                                pass
                            return len(processing_order)

                        items.sort(key=key_function)
                    for k, v in items:
                        self._process_internal(k, v, on_element)

            next(ret, '')

        except ReplaceElement as re:
            self.pop()
            popped = True
            for k, v in re.data.items():
                self._process_internal(k, v, on_element)
        finally:
            if not popped:
                self.pop()



from oarepo_model_builder.builders.jsonschema import JSONSchemaBuilder
from oarepo_model_builder.builders.mapping import MappingBuilder
from oarepo_model_builder.invenio.invenio_record_schema import InvenioRecordSchemaBuilder
from oarepo_model_builder.property_preprocessors import PropertyPreprocessor, process
from oarepo_model_builder.utils.deepmerge import deepmerge
from oarepo_model_builder.stack import ModelBuilderStack
from oarepo_model_builder.utils.schema import is_schema_element, match_schema, Ref


class TypeShortcutsPreprocessor(PropertyPreprocessor):
    TYPE = 'type_shortcuts'

    @process(model_builder='*', path='/model/**',
             condition=lambda current, stack: is_schema_element(stack))
    def modify_type_shortcuts(self, data, stack: ModelBuilderStack, **kwargs):
        schema_types = match_schema(stack)
        top = schema_types[-1]

        if not isinstance(top, Ref):
            return data

        self.set_type(top.element_type, data)
        self.set_child_arrays(top.element_type, data)

        return data

    def set_type(self, element_type, data):
        if 'type' in data:
            return

        if element_type in ('property', 'items'):
            if 'properties' in data:
                data['type'] = 'object'
            elif 'items' in data:
                data['type'] = 'array'

    def set_child_arrays(self, element_type, data):
        if element_type == 'properties':
            for k, v in list(data.items()):
                if k.endswith('[]'):
                    data.pop(k)
                    data[k[:-2]] = self.create_array(v)

    @staticmethod
    def create_array(value):
        ret = {}
        array = {}
        for k, v in value.items():
            if k.endswith('[]'):
                ret[k[:-2]] = v
            else:
                array[k] = v
        ret['type'] = 'array'
        ret['items'] = array
        return ret

    @process(model_builder='*', path='/model',
             condition=lambda current, stack: is_schema_element(stack))
    def add_model_type(self, data, stack: ModelBuilderStack, **kwargs):
        if 'type' not in data:
            data['type'] = 'object'
        return data



from oarepo_model_builder.builders.python import PythonBuilder
from oarepo_model_builder.outputs.python import PythonOutput
from oarepo_model_builder.utils.hyphen_munch import HyphenMunch
from oarepo_model_builder.utils.jinja import package_name


class InvenioBaseClassPythonBuilder(PythonBuilder):
    class_config = None
    template = None

    def finish(self, **extra_kwargs):
        python_path = self.class_to_path(self.settings.python[self.class_config])
        self.process_template(python_path, self.template,
                              current_package_name=package_name(self.settings.python[self.class_config]),
                              **extra_kwargs)

    def process_template(self, python_path, template, **extra_kwargs):
        self.create_parent_modules(python_path)
        output: PythonOutput = self.builder.get_output(
            'python',
            python_path
        )
        context = HyphenMunch(settings=self.settings, python=self.settings.python,
                              **extra_kwargs)
        template = self.call_components('invenio_before_python_template', template, context=context)
        output.merge(
            template,
            context
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oarepo_model_builder',
 'oarepo_model_builder.builders',
 'oarepo_model_builder.builtin_models',
 'oarepo_model_builder.invenio',
 'oarepo_model_builder.loaders',
 'oarepo_model_builder.model_preprocessors',
 'oarepo_model_builder.outputs',
 'oarepo_model_builder.property_preprocessors',
 'oarepo_model_builder.templates',
 'oarepo_model_builder.utils']

package_data = \
{'': ['*'], 'oarepo_model_builder.invenio': ['templates/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'black>=21.11b1,<22.0',
 'click>=7.1',
 'deepdiff>=5.6.0,<6.0.0',
 'isort>=5.10.1,<6.0.0',
 'json5>=0.9.6,<0.10.0',
 'jsonpointer>=2.2,<3.0',
 'libcst>=0.3.19,<0.4.0',
 'munch>=2.5.0,<3.0.0',
 'tomlkit>=0.7.2,<0.8.0']

entry_points = \
{'console_scripts': ['oarepo-compile-model = oarepo_model_builder.cli:run'],
 'oarepo.models': ['invenio = '
                   'oarepo_model_builder.builtin_models:invenio.json'],
 'oarepo_model_builder.builders': ['020-jsonschema = '
                                   'oarepo_model_builder.builders.jsonschema:JSONSchemaBuilder',
                                   '030-mapping = '
                                   'oarepo_model_builder.builders.mapping:MappingBuilder',
                                   '050-poetry = '
                                   'oarepo_model_builder.builders.poetry:PoetryBuilder',
                                   '100-python_structure = '
                                   'oarepo_model_builder.builders.python_structure:PythonStructureBuilder',
                                   '110-invenio_record = '
                                   'oarepo_model_builder.invenio.invenio_record:InvenioRecordBuilder',
                                   '120-invenio_record_metadata = '
                                   'oarepo_model_builder.invenio.invenio_record_metadata:InvenioRecordMetadataBuilder',
                                   '130-invenio_record_schema = '
                                   'oarepo_model_builder.invenio.invenio_record_schema:InvenioRecordSchemaBuilder',
                                   '200-invenio_record_permissions = '
                                   'oarepo_model_builder.invenio.invenio_record_permissions:InvenioRecordPermissionsBuilder',
                                   '300-invenio_record_search_options = '
                                   'oarepo_model_builder.invenio.invenio_record_search:InvenioRecordSearchOptionsBuilder',
                                   '310-invenio_record_service_config = '
                                   'oarepo_model_builder.invenio.invenio_record_service_config:InvenioRecordServiceConfigBuilder',
                                   '320-invenio_record_service = '
                                   'oarepo_model_builder.invenio.invenio_record_service:InvenioRecordServiceBuilder',
                                   '340-invenio_record_dumper = '
                                   'oarepo_model_builder.invenio.invenio_record_dumper:InvenioRecordDumperBuilder',
                                   '400-invenio_record_resource_config = '
                                   'oarepo_model_builder.invenio.invenio_record_resource_config:InvenioRecordResourceConfigBuilder',
                                   '410-invenio_record_resource = '
                                   'oarepo_model_builder.invenio.invenio_record_resource:InvenioRecordResourceBuilder',
                                   '420-invenio_views = '
                                   'oarepo_model_builder.invenio.invenio_views:InvenioViewsBuilder',
                                   '500-invenio_config = '
                                   'oarepo_model_builder.invenio.invenio_config:InvenioConfigBuilder',
                                   '600-invenio_ext = '
                                   'oarepo_model_builder.invenio.invenio_ext:InvenioExtBuilder',
                                   '610-invenio_ext_poetry = '
                                   'oarepo_model_builder.invenio.invenio_ext_poetry:InvenioExtPoetryBuilder',
                                   '700-invenio_ext = '
                                   'oarepo_model_builder.invenio.invenio_proxies:InvenioProxiesBuilder',
                                   '900-invenio_sample_app_poetry = '
                                   'oarepo_model_builder.invenio.invenio_sample_app_poetry:InvenioSampleAppPoetryBuilder',
                                   '910-invenio_record_metadata_alembic_poetry '
                                   '= '
                                   'oarepo_model_builder.invenio.invenio_record_metadata_alembic_poetry:InvenioRecordMetadataAlembicPoetryBuilder',
                                   '920-invenio_record_metadata_models_poetry '
                                   '= '
                                   'oarepo_model_builder.invenio.invenio_record_metadata_models_poetry:InvenioRecordMetadataModelsPoetryBuilder',
                                   '930-invenio_resource_poetry = '
                                   'oarepo_model_builder.invenio.invenio_record_resource_poetry:InvenioRecordResourcePoetryBuilder',
                                   '940-invenio_record_search_poetry = '
                                   'oarepo_model_builder.invenio.invenio_record_search_poetry:InvenioRecordSearchPoetryBuilder',
                                   '950-invenio_record_jsonschemas_poetry = '
                                   'oarepo_model_builder.invenio.invenio_record_jsonschemas_poetry:InvenioRecordJSONSchemasPoetryBuilder'],
 'oarepo_model_builder.loaders': ['json = '
                                  'oarepo_model_builder.loaders:json_loader',
                                  'json5 = '
                                  'oarepo_model_builder.loaders:json_loader',
                                  'yaml = '
                                  'oarepo_model_builder.loaders:yaml_loader',
                                  'yml = '
                                  'oarepo_model_builder.loaders:yaml_loader'],
 'oarepo_model_builder.model_preprocessors': ['01-default = '
                                              'oarepo_model_builder.model_preprocessors.default_values:DefaultValuesModelPreprocessor',
                                              '10-invenio = '
                                              'oarepo_model_builder.model_preprocessors.invenio:InvenioModelPreprocessor',
                                              '20-elasticsearch = '
                                              'oarepo_model_builder.model_preprocessors.elasticsearch:ElasticsearchModelPreprocessor'],
 'oarepo_model_builder.ouptuts': ['jsonschema = '
                                  'oarepo_model_builder.outputs.jsonschema:JSONSchemaOutput',
                                  'mapping = '
                                  'oarepo_model_builder.outputs.mapping:MappingOutput',
                                  'python = '
                                  'oarepo_model_builder.outputs.python:PythonOutput',
                                  'toml = '
                                  'oarepo_model_builder.outputs.toml:TOMLOutput'],
 'oarepo_model_builder.property_preprocessors': ['100-type_shortcuts = '
                                                 'oarepo_model_builder.property_preprocessors.type_shortcuts:TypeShortcutsPreprocessor',
                                                 '500-text_keyword = '
                                                 'oarepo_model_builder.property_preprocessors.text_keyword:TextKeywordPreprocessor',
                                                 '600-date = '
                                                 'oarepo_model_builder.property_preprocessors.date:DatePreprocessor'],
 'oarepo_model_builder.templates': ['99-base_templates = '
                                    'oarepo_model_builder.invenio']}

setup_kwargs = {
    'name': 'oarepo-model-builder',
    'version': '0.9.4',
    'description': 'An utility library that generates OARepo required data model files from a JSON specification file',
    'long_description': '# OARepo model builder\n\nA library and command-line tool to generate invenio model project from a single model file.\n\n<!--TOC-->\n\n- [OARepo model builder](#oarepo-model-builder)\n  - [CLI Usage](#cli-usage)\n    - [Installing model builder as dev dependency](#installing-model-builder-as-dev-dependency)\n    - [Installing model builder in a separate virtualenv](#installing-model-builder-in-a-separate-virtualenv)\n    - [Running model builder](#running-model-builder)\n  - [Model file](#model-file)\n    - [Model file structure](#model-file-structure)\n    - ["model" section](#model-section)\n    - ["settings" section](#settings-section)\n    - ["plugins" section](#plugins-section)\n  - [Builder as a library (using via API)](#builder-as-a-library-using-via-api)\n  - [Writing custom plugins](#writing-custom-plugins)\n\n<!--TOC-->\n\n## CLI Usage\n\nTo use the model builder client, you first have to install the model builder somewhere.\n\n### Installing model builder as dev dependency\n\nInitialize your new project with ``poetry create`` and then add model builder\nwith ``poetry add --dev oarepo-model-builder``. This is the simplest solution but has a disadvantage - as poetry always\ninstalls dev dependencies in build & test, but not in production, in your development environment you will have extra\npackages installed. If you happen to use them, you will break your production build.\n\n### Installing model builder in a separate virtualenv\n\nCreate a separate virtualenv and install model builder into it:\n\n```bash\npython3.10 -m venv .venv-builder\n(source .venv-builder/bin/activate; pip install -U pip setuptools wheel; pip install oarepo-model-builder)\n```\n\nThen for ease of use add the following aliases\n\n```bash\nalias oarepo-compile-model="$PWD/.venv-builder/bin/oarepo-compile-model"\nalias oarepo-model-builder-pip="$PWD/.venv-builder/bin/pip"\n```\n\nUse the ``oarepo-model-builder-pip`` if you need to install plugins to the model builder.\n\n### Running model builder\n\n```bash\noarepo-compile-model model.yaml\n```\n\nwill compile the model.yaml into the current directory. Options:\n\n```bash\n  --output-directory <dir> Output directory where the generated files will be\n                           placed. Defaults to "."\n  --package <name>         Package into which the model is generated. If not\n                           passed, the name of the current directory,\n                           converted into python package name, is used.\n  --set <name=value>       Overwrite option in the model file. \n                           Example --set settings.elasticsearch.keyword-ignore-above=20\n  -v                       Increase the verbosity. This option can be used\n                           multiple times.\n  --config <filename>      Load a config file and replace parts of the model\n                           with it. The config file can be a json, yaml or a\n                           python file. If it is a python file, it is\n                           evaluated with the current model stored in the\n                           "oarepo_model" global variable and after the\n                           evaluation all globals are set on the model.\n  --isort / --skip-isort   Call isort on generated sources (default: yes)\n  --black / --skip-black   Call black on generated sources (default: yes)\n```\n\n## Model file\n\nA model is a json/yaml file including description of the model and processing settings. Example:\n\n```yaml\nversion: 1.0.0\nmodel:\n  properties:\n    title:\n      type: fulltext+keyword\n      oarepo:ui:\n        label: Title\n      oarepo:mapping:\n      # anything in here will be put into the mapping file\n      # fulltext+keyword type automatically creates "type: text" \n      # with subfield \'keyword\' of type keyword\nsettings:\n  package: uct.titled_model \n```\n\n### Model file structure\n\nA model is a json/yaml file with the following structure:\n\n```yaml\nversion: 1.0.0\nmodel:\n  properties:\n    title:\n      type: fulltext+keyword\nsettings:\n  <generic settings here>\n  python: ...\n  elasticsearch: ...\nplugins: ...\n```\n\nThere might be more sections (documentation etc.), but only the ``settings``, ``model`` and ``plugins``\nare currently processed.\n\n### "model" section\n\nThis section is described in [model.md](docs/model.md)\n\n### "settings" section\n\nThe settings section contains various configuration settings. In most cases you want to set only the `package` option as\nin above because all other settings are derived from it. Even the `package`\noption might be omitted - in this case the package name will be the last component of the output directory (with dashes\nconverted to underscores).\n\nThe rest of the settings are described in [model-generic-settings.md](docs/model-generic-settings.md)\n\nAdvanced use cases might require to modify [the python settings](docs/model-python-settings.md) or\n[elasticsearch settings](docs/model-elasticsearch-settings.md) (for example, to define custom analyzers).\n\n### "plugins" section\n\nSee [plugins and the processing order](docs/model-plugins.md) for details.\n\n## Builder as a library (using via API)\n\nTo invoke the builder programmatically, see [using the API](docs/using-api.md).\n\n## Writing custom plugins\n\nSee [writing plugins](docs/extending-api.md) if you want to extend the building process with your own plugins.\n',
    'author': 'Miroslav Bauer',
    'author_email': 'bauer@cesnet.cz',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordServiceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_service_config'
    class_config = 'record-service-config-class'
    template = 'record-service-config'



from .invenio_base import InvenioBaseClassPythonBuilder


class InvenioRecordResourceConfigBuilder(InvenioBaseClassPythonBuilder):
    TYPE = 'invenio_record_resource_config'
    class_config = 'record-resource-config-class'
    template = 'record-resource-config'



from pathlib import Path


class AbstractFileSystem:
    def open(self, path, *args, **kwargs):
        raise Exception('Not implemented')

    def exists(self, path):
        raise Exception('Not implemented')

    def mkdir(self, path):
        raise Exception('Not implemented')


class FileSystem(AbstractFileSystem):
    def open(self, path, *args, **kwargs):
        return open(path, *args, **kwargs)

    def exists(self, path):
        return Path(path).exists()

    def mkdir(self, path):
        Path(path).mkdir(exist_ok=True, parents=True)
